/**
 * @file c2nload.c
 * The main program
 * @author Marko Mkel <msmakela@nic.funet.fi>
 */

/* Copyright  2001,2002 Marko Mkel.

   This file is part of C2NLOAD, a program for loading files to
   Commodore 8-bit computers via the C2N232 device.

   C2NLOAD is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2NLOAD is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

/** Program version */
#define VERSION "1.0.2"
/** Release date (day.month.year) */
#define DATE "12.12.2002"
/** Name of the program */
#define PROG "c2nload"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>

/** The name of the serial device */
static const char* serialdev;

#if defined __WIN32 || defined __WIN32__
# ifndef WIN32
#  define WIN32
# endif
# include <windows.h>

/** file handle for the serial device (0=none) */
static HANDLE serialfd = 0;
/** previous terminal input/output settings for the serial line */
static DCB tio;
/** previous timeout settings for the serial line */
static COMMTIMEOUTS tio2;
/** new terminal input/output settings for the serial line */
static DCB newtio;
/** new timeout settings for the serial line */
static COMMTIMEOUTS newtio2;

/** Report an error for accessing the serial device
 * @param op	the failed operation
 */
static void
report_error (const char* op)
{
  char* msg;
  long err = GetLastError ();
  FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		 FORMAT_MESSAGE_FROM_SYSTEM | 
		 FORMAT_MESSAGE_IGNORE_INSERTS,
		 0,
		 err,
		 0, /* default language */
		 (LPTSTR) &msg,
		 0,
		 0);
  AnsiToOem (msg, msg);
  fputs (PROG " -c ", stderr);
  fputs (serialdev, stderr);
  fputs (": ", stderr);
  fputs (op, stderr);
  fprintf (stderr, " failed with code %ld: ", err);
  fputs (msg, stderr);
  LocalFree (msg);
}
#elif defined __AMIGA__
static const char version[] = "$VER: C2N " VERSION " (" DATE ")";
# include <proto/exec.h>
# include <devices/serial.h>
# include <devices/timer.h>
# include <dos/dos.h>
# include <clib/alib_protos.h>
# include <clib/dos_protos.h>
/** extended input/output request structure */
static struct IOExtSer ser;
/** timer request */
static struct timerequest* timer;
/** dummy serial file descriptor */
# define serialfd !!ser.IOSer.io_Device

/** Report an error for accessing the serial device
 * @param op	the failed operation
 */
static void
report_error (const char* op)
{
  fputs (PROG " -c ", stderr);
  fputs (serialdev, stderr);
  fputs (": ", stderr);
  fputs (op, stderr);
  fprintf (stderr, " returned %d\n", ser.IOSer.io_Error);
}
#else /* unix */
# include <termios.h>
# include <unistd.h>
# include <fcntl.h>

/** file descriptor for the serial device (0=none) */
static int serialfd = 0;
/** previous terminal input/output settings for the serial line */
static struct termios tio;
/** new terminal input/output settings for the serial line */
static struct termios newtio;

/** Report an error for accessing the serial device
 * @param op	the failed operation
 */
static void
report_error (const char* op)
{
  fputs (PROG " -c ", stderr);
  fputs (serialdev, stderr);
  fputs (": ", stderr);
  perror (op);
}
#endif

#include "c2nload.h"
#include "encode.h"
#include "models.h"

/** Cleanup function */
static void
cleanup (void)
{
#ifdef WIN32
  if (serialfd && serialfd != INVALID_HANDLE_VALUE) {
    PurgeComm (serialfd, PURGE_TXCLEAR | PURGE_RXCLEAR);
    SetCommState (serialfd, &tio);
    SetCommTimeouts (serialfd, &tio2);
    SetCommBreak (serialfd);
    Sleep (20); /* send break signal for at least 20 milliseconds */
    ClearCommBreak (serialfd);
    CloseHandle (serialfd);
    serialfd = 0;
  }
#elif defined __AMIGA__
  if (ser.IOSer.io_Device && ser.IOSer.io_Message.mn_ReplyPort) {
    ser.IOSer.io_Command = CMD_CLEAR;
    if (DoIO ((struct IORequest*) &ser))
      fprintf (stderr, "cleanup: DoIO (CMD_CLEAR) returned %d\n",
	       ser.IOSer.io_Error);
    else {
      ser.IOSer.io_Command = SDCMD_BREAK;
      if (DoIO ((struct IORequest*) &ser))
	fprintf (stderr, "cleanup: DoIO (SDCMD_BREAK) returned %d\n",
		 ser.IOSer.io_Error);
    }
  }
  if (ser.IOSer.io_Device)
    CloseDevice ((struct IORequest*) &ser);
  if (ser.IOSer.io_Message.mn_ReplyPort)
    DeletePort (ser.IOSer.io_Message.mn_ReplyPort);
  memset (&ser, 0, sizeof ser);
  if (timer) {
    if (timer->tr_node.io_Message.mn_ReplyPort)
      DeletePort (timer->tr_node.io_Message.mn_ReplyPort);
    CloseDevice ((struct IORequest*) timer);
    DeleteExtIO ((struct IORequest*) timer);
    timer = 0;
  }
#else /* unix */
  if (serialfd > 0) {
    tcflush (serialfd, TCIOFLUSH);
    tcsetattr (serialfd, TCSANOW, &tio);
    tcsendbreak (serialfd, 0);
    close (serialfd);
    serialfd = 0;
  }
#endif
}

/** Signal handler
 * @param num	the signal number
 */
static void
sig (int num)
{
  cleanup ();
  signal (num, SIG_DFL);
  raise (num);
}

/** buffer for the pulse stream */
static char buf[32768];

/** number of bytes read from or written to buf */
static unsigned datain = 0;

#ifdef WIN32
/** Read data from the serial line
 * @param buf		the data to be received
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
static int
serdev_read (void* buf,
	     unsigned len)
{
  unsigned long length;
  if (ReadFile (serialfd, buf, len, &length, 0))
    return 0;
  report_error ("ReadFile");
  return 1;
}

/** Write data to the serial line
 * @param buf	data to be written
 * @param len	length of the data
 * @return	zero on success, nonzero on failure
 */
static int
serdev_write (const void* buf,
	      unsigned len)
{
  unsigned long length;
  if (WriteFile (serialfd, buf, len, &length, 0)) {
    if (FlushFileBuffers (serialfd))
      return 0;
    report_error ("FlushFileBuffers");
  }
  else
    report_error ("WriteFile");
  return 1;
}
#endif /* WIN32 */
#ifdef __AMIGA__
/** Flush the input/output operation on the serial line
 * @return	the WaitIO exit status
 */
static BYTE
flush_serial (void)
{
  const long waitsigs = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D |
    1L << ser.IOSer.io_Message.mn_ReplyPort->mp_SigBit;
  for (;;) {
    if (CheckIO ((struct IORequest*) &ser))
      return WaitIO ((struct IORequest*) &ser);
    if (Wait (waitsigs) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D)) {
      AbortIO ((struct IORequest*) &ser);
      WaitIO ((struct IORequest*) &ser);
      raise (SIGINT);
    }
  }
}

/** Read data from the serial line
 * @param buf		the data to be received
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
static int
serdev_read (void* buf,
	     unsigned len)
{
  ser.IOSer.io_Command = CMD_READ;
  ser.IOSer.io_Length = len;
  ser.IOSer.io_Data = buf;
  SendIO ((struct IORequest*) &ser);
  if (flush_serial ())
    report_error ("WaitIO (CMD_READ)");
  return ser.IOSer.io_Error;
}

/** Write data to the serial line
 * @param buf	data to be written
 * @param len	length of the data
 * @return	zero on success, nonzero on failure
 */
static int
serdev_write (const void* buf,
	      unsigned len)
{
  ser.IOSer.io_Command = CMD_WRITE;
  ser.IOSer.io_Length = len;
  ser.IOSer.io_Data = (void*) buf;
  SendIO ((struct IORequest*) &ser);
  if (flush_serial ())
    report_error ("WaitIO (CMD_WRITE)");
  return ser.IOSer.io_Error;
}
#endif /* AMIGA */

/** Switch the C2N232 to idle mode
 * @return	zero on success, nonzero on failure
 */
static int
go_idle (void)
{
#ifdef WIN32
  unsigned long len;
  /* send a NUL character and await a response */
  *buf = 0;
  if (!FlushFileBuffers (serialfd))
    report_error ("FlushFileBuffers");
  else if (Sleep (50), !WriteFile (serialfd, "", 1, &len, 0))
    report_error ("WriteFile");
  else if (!ReadFile (serialfd, buf, sizeof buf, &len, 0))
    report_error ("ReadFile");
  else if (!len || buf[len - 1] != '0') {
    fputs (PROG " -c ", stderr);
    fputs (serialdev, stderr);
    fputs (len
	   ? ": unexpected response from C2N232 device\n"
	   : ": no response from C2N232 device\n",
	   stderr);
  }
  else
    return 0;
  return 1;
#elif defined __AMIGA__
  Delay (3); /* wait 3 * 20 milliseconds */
  /* send a NUL character and await a response */
  *buf = 0;
  if (!serdev_write (buf, 1)) {
    /* await a response */
    const long waitsigs = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D |
      1L << timer->tr_node.io_Message.mn_ReplyPort->mp_SigBit |
      1L << ser.IOSer.io_Message.mn_ReplyPort->mp_SigBit;
    ser.IOSer.io_Command = CMD_READ;
    ser.IOSer.io_Length = sizeof buf;
    ser.IOSer.io_Data = buf;
    SendIO ((struct IORequest*) &ser);
    timer->tr_node.io_Command = TR_ADDREQUEST;
    timer->tr_time.tv_secs = 0;
    timer->tr_time.tv_micro = 500000L;
    SendIO ((struct IORequest*) timer);
    for (;;) {
      if (Wait (waitsigs) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D)) {
	AbortIO ((struct IORequest*) timer);
	WaitIO ((struct IORequest*) timer);
	AbortIO ((struct IORequest*) &ser);
	WaitIO ((struct IORequest*) &ser);
	raise (SIGINT);
      }
      if (CheckIO ((struct IORequest*) &ser)) {
	AbortIO ((struct IORequest*) timer);
	WaitIO ((struct IORequest*) timer);
	if (WaitIO ((struct IORequest*) &ser))
	  report_error ("WaitIO (CMD_READ)");
	else {
	got_data:
	  if (buf[ser.IOSer.io_Actual - 1] == '0')
	    return 0;
	  else {
	    fputs (PROG " -c ", stderr);
	    fputs (serialdev, stderr);
	    fputs (": unexpected response from C2N232 device\n", stderr);
	  }
	}
	break;
      }
      else if (CheckIO ((struct IORequest*) timer)) {
	WaitIO ((struct IORequest*) timer);
	AbortIO ((struct IORequest*) &ser);
	WaitIO ((struct IORequest*) &ser);
	if (ser.IOSer.io_Actual > 0)
	  goto got_data;
	fputs (PROG " -c ", stderr);
	fputs (serialdev, stderr);
	fputs (": no response from C2N232 device\n", stderr);
	return 1;
      }
    }
  }
  return 1;
#else /* unix */
  char last = 0;
  newtio.c_cc[VMIN] = 0; /* non-blocking read */
  newtio.c_cc[VTIME] = 1; /* 1/10th second timeout */
  if (tcsetattr (serialfd, TCSANOW, &newtio))
    report_error ("tcsetattr");
  else if (tcdrain (serialfd))
    report_error ("tcdrain");
  else if (usleep (50000),
	   1 != write (serialfd, &last, 1))
    report_error ("write");
  else
    for (;;) {
      int len = read (serialfd, buf, sizeof buf);
      if (!len) {
	if (last == '0') {
	  newtio.c_cc[VMIN] = 1; /* blocking read */
	  newtio.c_cc[VTIME] = 0; /* no timeout */
	  if (!tcsetattr (serialfd, TCSANOW, &newtio))
	    return 0;
	  report_error ("tcsetattr");
	  return 2;
	}
	if (last || !errno) {
	  fputs (PROG " -c ", stderr);
	  fputs (serialdev, stderr);
	  fputs (last
		 ? ": unexpected response from C2N232 device\n"
		 : ": no response from C2N232 device\n", stderr);
	}
	else
	  report_error ("read");
	break;
      }
      last = buf[len - 1];
    }
  return 1;
#endif
}

/** Write a pulse to the serial line
 * @param p	the pulse to be written
 */
void
pulsewrite (enum pulse p)
{
  if (datain < sizeof buf) {
    buf[datain++] = p;
    if (datain == sizeof buf)
      fputs ("pulsewrite: truncating the pulse stream\n", stderr);
  }
}

/** The main function
 * @param argc	argument count
 * @param argv	argument vector
 * @return	0 if successful
 */
int
main (int argc, char** argv)
{
  /** program name */
  const char* prog = *argv;
  /** currently selected Commodore model */
  const struct model* model = 0;
  /** load operation */
  static const char op_load[5] = { 3, (char) 255, 22, 31, 41 };
  /** load operation for the 264 series */
  static const char op_load264[5] = { 3, (char) 255, 30, 60, 120 };

  /** initial sync pulse count */
  static unsigned sync_begin = 0;
  /** intra-block sync pulse count */
  static unsigned sync_intra = 0;
  /** jump address */
  unsigned jump = 0;
  /** flag: quiet mode */
  int quiet = 0;

  atexit (cleanup);
  signal (SIGINT, sig);

  /* process the options */
  for (argv++; argc > 1 && **argv == '-'; argv++, argc--) {
    const char* opts = *argv;

    if (!strcmp (opts, "--")) { /* disable processing further options */
      argv++;
      argc--;
      break;
    }

    while (*++opts) /* process all flags */
      switch (*opts) {
      case 'c':
	if (argc <= 2) {
	  fputs (PROG " -c: missing argument\n", stderr);
	  goto Usage;
	}
	else if (serialfd) {
	  fputs (PROG " -c: got already -c ", stderr);
	  fputs (serialdev, stderr);
	  putc ('\n', stderr);
	  goto Usage;
	}
	serialdev = *++argv;
#ifdef WIN32
	do {
	  if ((serialfd = CreateFile (serialdev,
				      GENERIC_WRITE | GENERIC_READ,
				      0, 0, OPEN_EXISTING, 0, 0)) ==
	      INVALID_HANDLE_VALUE)
	    report_error ("CreateFile");
	  else if (!GetCommState (serialfd, &tio))
	    report_error ("GetCommState");
	  else if (!GetCommTimeouts (serialfd, &tio2))
	    report_error ("GetCommTimeouts");
	  else
	    continue;
	  return 2;
	} while (0);

	memcpy (&newtio, &tio, sizeof newtio);
	memcpy (&newtio2, &tio2, sizeof newtio2);
	newtio.BaudRate = CBR_38400;
	newtio.fBinary = 1;
	newtio.fParity = newtio.fOutxCtsFlow = newtio.fOutxDsrFlow = 0;
	newtio.fDtrControl = DTR_CONTROL_DISABLE;
	newtio.fDsrSensitivity = 0;
	newtio.fOutX = 1, newtio.fInX = 0;
	newtio.fErrorChar = newtio.fNull = 0;
	newtio.fRtsControl = RTS_CONTROL_DISABLE;
	newtio.fAbortOnError = 0;
	newtio.ByteSize = 8;
	newtio.Parity = NOPARITY;
	newtio.StopBits = ONESTOPBIT;
	newtio.XonChar = 021; /* DC1, ctrl-q */
	newtio.XoffChar = 023; /* DC3, ctrl-s */

	newtio2.ReadIntervalTimeout = 100; /* 1/10th second */
	newtio2.ReadTotalTimeoutConstant = 100; /* 1/10th second */
	newtio2.ReadTotalTimeoutMultiplier =
	  newtio2.WriteTotalTimeoutMultiplier =
	  newtio2.WriteTotalTimeoutConstant = 0;

	do {
	  if (!SetCommState (serialfd, &newtio))
	    report_error ("SetCommState");
	  else if (!SetCommTimeouts (serialfd, &newtio2))
	    report_error ("SetCommTimeouts");
	  else
	    continue;
	  return 2;
	} while (0);
#elif defined __AMIGA__
	memset (&ser, 0, sizeof ser);
	do {
	  if (!(ser.IOSer.io_Message.mn_ReplyPort =
		(struct MsgPort*) CreatePort (0, 0))) {
	  createport:
	    fputs ("CreatePort failed\n", stderr);
	  }
	  else if (OpenDevice ((const unsigned char*) serialdev, 0,
			       (struct IORequest*) &ser, 0))
	    report_error ("OpenDevice");
	  else {
	    struct MsgPort* timerport = CreatePort (0, 0);
	    if (!timerport)
	      goto createport;
	    else if (!(timer = (struct timerequest*)
		       CreateExtIO (timerport, sizeof timer))) {
	      DeletePort (timerport);
	      fputs ("CreateExtIO failed\n", stderr);
	    }
	    else if (OpenDevice ((const unsigned char*) "timer.device",
				 UNIT_VBLANK, (struct IORequest*) timer, 0L))
	      fprintf (stderr, "OpenDevice for timer.device returned %d\n",
		       timer->tr_node.io_Error);
	    else
	      continue;
	  }
	  return 2;
	} while (0);
	ser.IOSer.io_Command = SDCMD_SETPARAMS;
	ser.io_Baud = 38400;
	ser.io_ExtFlags = 0;
	ser.io_RBufLen = 2 * sizeof buf;
	ser.io_ReadLen = ser.io_WriteLen = 8;
	ser.io_StopBits = 1;
	ser.io_SerFlags = 0;
	if (DoIO ((struct IORequest*) &ser)) {
	  report_error ("DoIO (SDCMD_SETPARAMS)");
	  return 2;
	}
#else /* unix */
	if ((serialfd = open (serialdev, O_RDWR | O_NONBLOCK | O_NOCTTY))
	    < 0) {
	  report_error ("open");
	  return 2;
	}
	else {
	  int flags = fcntl (serialfd, F_GETFL);
	  if (flags < 0) {
	    report_error ("fcntl(GETFL)");
	    close (serialfd);
	    serialfd = -1;
	    return 2;
	  }
	  else if (flags & O_NONBLOCK &&
		   fcntl (serialfd, F_SETFL, flags & ~O_NONBLOCK) < 0) {
	    report_error ("fcntl(SETFL)");
	    close (serialfd);
	    serialfd = -1;
	    return 2;
	  }
	}

	if (tcgetattr (serialfd, &tio)) {
	  report_error ("tcgetattr");
	  close (serialfd);
	  serialfd = -1;
	  return 2;
	}

	memcpy (&newtio, &tio, sizeof newtio);
	newtio.c_iflag = IGNBRK | IGNPAR | IXON;
	newtio.c_oflag = 0;
	newtio.c_cflag = CS8 | CLOCAL | CREAD;
	newtio.c_lflag = 0;
	memset (newtio.c_cc, 0, sizeof newtio.c_cc);
	newtio.c_cc[VSTART] = 021; /* DC1, ctrl-q */
	newtio.c_cc[VSTOP] = 023; /* DC3, ctrl-s */
	cfsetispeed (&newtio, B38400), cfsetospeed (&newtio, B38400);
	tcflush (serialfd, TCIOFLUSH);
	tcsendbreak (serialfd, 0);
#endif
	if (go_idle ())
	  return 2;
	argc--;
	break;
      case 'm':
	if (argc <= 2) {
	  fputs (PROG " -m: missing argument\n", stderr);
	  goto Usage;
	}
	else {
	  argv++; argc--;
	  for (model = models; model->name; model++)
	    if (!strcmp (*argv, model->name))
	      break;
	  if (model->name)
	    break;
	  fputs (PROG " -m: instead of ", stderr);
	  fputs (*argv, stderr);
	  fputs (", try ", stderr);
	  for (model = models;; ) {
	    fputs (model->name, stderr);
	    /* skip the aliases for the model */
	    while (model[1].loader == model->loader) model++;
	    if (!(++model)->name) {
	      putc ('\n', stderr);
	      break;
	    }
	    putc (' ', stderr);
	  }
	  model = 0;
	  return 1;
	}
	break;
      case 'h':
	goto Usage;
      case 'j':
	if (argc <= 2) {
	  fputs (PROG " -j: missing argument\n", stderr);
	  goto Usage;
	}
	else {
	  char* r;
	  char* s = *++argv;
	  argc--;
	  jump = strtoul (s, &r, 0);
	  if (!*s || *r) {
	    fprintf (stderr, PROG " -j: address expected, got %s\n", *argv);
	    goto Usage;
	  }
	}
	break;
      case 's':
	if (argc <= 2) {
	  fputs (PROG " -s: missing argument\n", stderr);
	  goto Usage;
	}
	else {
	  char* r;
	  char* s = *++argv;
	  argc--;
	  sync_begin = strtoul (s, &r, 0);
	  if (*r != ',') {
	  sUsage:
	    fputs (PROG " -s: begin,intra expected, got ", stderr);
	    fputs (*argv, stderr);
	    putc ('\n', stderr);
	    goto Usage;
	  }
	  s = r + 1;
	  sync_intra = strtoul (s, &r, 0);
	  if (*r)
	    goto sUsage;
	}
	break;
      case 'q':
	quiet = 1;
	break;

      default:
	fputs (PROG " -", stderr);
	putc (*opts, stderr);
	fputs (": unknown option\n", stderr);
	goto Usage;
      }
  }

  if (argc < 2 || !serialfd || !model) {
    if (!serialfd)
      fputs (PROG ": missing -c option\n", stderr);
    if (!model)
      fputs (PROG ": missing -m option\n", stderr);
    if (argc < 2)
      fputs (PROG ": file argument missing\n", stderr);
  Usage:
    fputs ("C2NLOAD " VERSION
	   " - Commodore C2N232 loader\n"
	   "Usage: ", stderr);
    fputs (prog, stderr);
    fputs (" [options] file(s)\n"
	   "Options: -c device: specify the name of the C2N232 device\n"
	   "         -m model: specify the model of the Commodore computer\n"
	   "         -h: request this help text\n"
	   "         -j address: jump to the address after loading\n"
	   "         -s begin,intra: specify lengths of sync pulse streams\n"
	   "         -q: be quiet\n"
	   "         --: Stop processing any further options.\n",
	   stderr);
    return 1;
  }

  /* Process the files. */

  for (; --argc; argv++) {
    /** start and length of the program */
    unsigned start, proglen;
    /** program data */
    static char progbuf[65536];
    {
      FILE* df;
      if (!(df = strcmp (*argv, "-")
	    ? fopen (*argv, "rb")
	    : stdin)) {
	fputs (*argv, stderr);
	perror (": fopen (read)");
	return 2;
      }
      start = getc (df);
      start |= getc (df) << 8;
      proglen = fread (progbuf, 1, sizeof progbuf, df);
      if (!proglen || !feof (df)) {
	fputs (*argv, stderr);
	fputs (": not a program file\n", stderr);
	if (df != stdin) fclose (df);
	return 2;
      }
      else if (ferror (df)) {
	fputs (*argv, stderr);
	perror (": fread");
	if (df != stdin) fclose (df);
	return 2;
      }
      if (df != stdin) fclose (df);
    }

#if defined WIN32 || defined __AMIGA__
    if (serdev_write (model->algo == a264
		      ? op_load264
		      : op_load, sizeof op_load) ||
	serdev_read (buf, 1))
      return 2;
#else /* unix */
    if (sizeof op_load != write (serialfd, model->algo == a264
				 ? op_load264
				 : op_load, sizeof op_load)) {
      report_error ("write");
      return 2;
    }

    if (!read (serialfd, buf, 1)) {
      if (errno)
	report_error ("read");
      else {
	fputs (PROG " -c ", stderr);
	fputs (serialdev, stderr);
	fputs (": no response from C2N232 device\n", stderr);
      }
      return 2;
    }
#endif

    if (*buf != '3') {
      fprintf (stderr, PROG " -c %s: C2N232 responded 0x%02x; expected '3'\n",
	       serialdev, (unsigned char) *buf);
      return 2;
    }

    encode (model->loader,
	    sync_begin ? sync_begin : 1500,
	    sync_intra ? sync_intra : 2000,
	    model->algo == a264,
	    *argv, start, proglen, jump);

    if (!quiet) {
      switch (model->algo) {
      case aBLANK:
      case a264:
	fputs (PROG ": Please hold down the Commodore key on the ", stderr);
	fputs (model->name, stderr);
	fputs ("!\n", stderr);
      case aBASIC:
      case aNORMAL:
	break;
      }
    }

#if defined WIN32 || defined __AMIGA__
    if (serdev_write (buf, datain))
      return 2;
#else /* unix */
    if (datain != write (serialfd, buf, datain))
      return report_error ("write"), 2;
#endif
    datain = 0;

    if (go_idle ())
      return 2;

    if (!quiet) {
      switch (model->algo) {
      case aBASIC:
	fputs (PROG ": Please RUN the loader on the ", stderr);
	fputs (model->name, stderr);
	fputs ("!\n", stderr);
	break;
      case aBLANK:
      case a264:
	fputs (PROG ": You may now release the Commodore key.\n", stderr);
      case aNORMAL:
	fputs (PROG ": loader started, loading the file.\n", stderr);
	break;
      }
    }

    buf[0] = 4;
    buf[1] = proglen - 1;
    buf[2] = (proglen - 1) >> 8;
#if defined WIN32 || defined __AMIGA__
    if (serdev_write (buf, 3) ||
	serdev_read (buf, 1))
      return 2;
#else /* unix */
    if (3 != write (serialfd, buf, 3))
      return report_error ("write"), 2;
    else if (1 != read (serialfd, buf, 1))
      return report_error ("read"), 2;
#endif
    if (*buf != '4') {
      fprintf (stderr, PROG " -c %s: C2N232 responded 0x%02x; expected '4'\n",
	       serialdev, (unsigned char) *buf);
      return 2;
    }
#if defined WIN32 || defined __AMIGA__
    if (serdev_write (progbuf, proglen) ||
	serdev_read (buf, 1))
      return 2;
#else /* unix */
    do {
      if (proglen != write (serialfd, progbuf, proglen))
	report_error ("write");
      else if (1 != read (serialfd, buf, 1)) {
	if (errno)
	  report_error ("read");
	else {
	  fputs (PROG " -c ", stderr);
	  fputs (serialdev, stderr);
	  fputs (": C2N232 timeout\n", stderr);
	}
      }
      else
	continue;
      return 2;
    } while (0);
#endif
    if (*buf != '@') {
      fprintf (stderr, PROG " -c %s: C2N232 responded 0x%02x; expected '@'\n",
	       serialdev, (unsigned char) *buf);
      return 2;
    }
  }

  return 0;
}
